@@ -12,8 +12,8 @@ use core::iter::FromIterator;
 ///
 /// This trait is not object safe, so can't be used dynamically. However it is possible to
 /// define a generic function over types implementing RingBuffer.
-pub trait RingBuffer<T: 'static + Default>:
-    Default + Index<isize, Output = T> + IndexMut<isize> + FromIterator<T> + crate::private::Sealed
+pub trait RingBuffer<T: 'static>:
+    Default + Index<isize, Output = T> + IndexMut<isize> + FromIterator<T>
 {
     /// Returns the length of the internal buffer.
     /// This length grows up to the capacity and then stops growing.
@@ -48,22 +48,7 @@ pub trait RingBuffer<T: 'static + Default>:
 
     /// Gets a value relative to the current index mutably. 0 is the next index to be written to with push.
     /// -1 and down are the last elements pushed and 0 and up are the items that were pushed the longest ago.
-    ///
-    /// # Safety
-    /// get_mut_impl is used to implement [`iter_mut`](Self::iter_mut). It requires that for indices in the range
-    /// 0..self.len(), every reference ***MUST ONLY BE RETURNED ONCE***. Any sane implementation of
-    /// get_mut does this anyway, but failing to do so results in the possibility to have multiple multiple
-    /// references to data inside the ringbuffer. (as per issue #25) This function is unsafe precisely because
-    /// of this issue.
-    unsafe fn get_mut_impl(&mut self, index: isize) -> Option<&mut T>;
-
-    /// Gets a value relative to the current index mutably. 0 is the next index to be written to with push.
-    /// -1 and down are the last elements pushed and 0 and up are the items that were pushed the longest ago.
-    fn get_mut(&mut self, index: isize) -> Option<&mut T> {
-        // Safety: calling get_mut_impl is not unsafe, implementing it might be unsafe because of
-        // the behaviour of iter_mut (see [`Self::get_mut_impl`])
-        unsafe { self.get_mut_impl(index) }
-    }
+    fn get_mut(&mut self, index: isize) -> Option<&mut T>;
 
     /// Gets a value relative to the start of the array (rarely useful, usually you want [`Self::get`])
     fn get_absolute(&self, index: usize) -> Option<&T>;
@@ -72,7 +57,7 @@ pub trait RingBuffer<T: 'static + Default>:
     fn get_absolute_mut(&mut self, index: usize) -> Option<&mut T>;
 
     /// Pushes a value onto the buffer. Cycles around if capacity is reached.
-    fn push(&mut self, e: T);
+    fn push(&mut self, value: T);
 
     /// Returns the value at the current index.
     /// This is the value that will be overwritten by the next push and also the value pushed
@@ -145,6 +130,26 @@ pub trait RingBuffer<T: 'static + Default>:
     {
         self.iter().any(|i| i == elem)
     }
+
+    /// Dequeues the top item off the ringbuffer. Returns a reference to the item. This means
+    /// that lifetimes will be problematic because as long as this reference exists,
+    /// you can not push to the queue. To solve this, use the pop method. This requires
+    /// the item to be clone. Easily moving out of the ringbuffer is sadly impossible.
+    ///
+    /// Returns None when the ringbuffer is empty.
+    fn dequeue_ref(&mut self) -> Option<&T>;
+
+    /// Dequeues the top item off the ringbuffer and returns an owned version. See the [`pop_ref`](Self::pop_ref) docs
+    fn dequeue(&mut self) -> Option<T>
+    where
+        T: Clone,
+    {
+        self.dequeue_ref().cloned()
+    }
+
+    /// Pops the top item off the queue, but does not return it. Instead it is dropped.
+    /// If the ringbuffer is empty, this function is a nop.
+    fn skip(&mut self);
 }
 
 mod iter {
@@ -153,13 +158,13 @@ mod iter {
 
     /// RingBufferIterator holds a reference to a RingBuffer and iterates over it. `index` is the
     /// current iterator position.
-    pub struct RingBufferIterator<'rb, T: 'static + Default, RB: RingBuffer<T>> {
+    pub struct RingBufferIterator<'rb, T: 'static, RB: RingBuffer<T>> {
         obj: &'rb RB,
         index: usize,
         phantom: PhantomData<T>,
     }
 
-    impl<'rb, T: 'static + Default, RB: RingBuffer<T>> RingBufferIterator<'rb, T, RB> {
+    impl<'rb, T: 'static, RB: RingBuffer<T>> RingBufferIterator<'rb, T, RB> {
         #[inline]
         pub fn new(obj: &'rb RB) -> Self {
             Self {
@@ -170,7 +175,7 @@ mod iter {
         }
     }
 
-    impl<'rb, T: 'static + Default, RB: RingBuffer<T>> Iterator for RingBufferIterator<'rb, T, RB> {
+    impl<'rb, T: 'static, RB: RingBuffer<T>> Iterator for RingBufferIterator<'rb, T, RB> {
         type Item = &'rb T;
 
         #[inline]
@@ -190,13 +195,13 @@ mod iter {
     ///
     /// WARNING: NEVER ACCESS THE `obj` FIELD. it's private on purpose, and can technically be accessed
     /// in the same module. However, this breaks the safety of `next()`
-    pub struct RingBufferMutIterator<'rb, T: 'static + Default, RB: RingBuffer<T>> {
+    pub struct RingBufferMutIterator<'rb, T: 'static, RB: RingBuffer<T>> {
         obj: &'rb mut RB,
         index: usize,
         phantom: PhantomData<T>,
     }
 
-    impl<'rb, T: 'static + Default, RB: RingBuffer<T>> RingBufferMutIterator<'rb, T, RB> {
+    impl<'rb, T: 'static, RB: RingBuffer<T>> RingBufferMutIterator<'rb, T, RB> {
         #[inline]
         pub fn new(obj: &'rb mut RB) -> Self {
             Self {
@@ -205,25 +210,13 @@ mod iter {
                 phantom: Default::default(),
             }
         }
-    }
 
-    impl<'rb, T: 'static + Default, RB: RingBuffer<T>> Iterator for RingBufferMutIterator<'rb, T, RB> {
-        type Item = &'rb mut T;
-
-        #[inline]
-        fn next(&mut self) -> Option<Self::Item> {
+        pub fn next(&mut self) -> Option<&mut T> {
             if self.index < self.obj.len() {
-                let res: Option<&'_ mut T> = unsafe { self.obj.get_mut_impl(self.index as isize) };
+                let res = self.obj.get_mut(self.index as isize);
                 self.index += 1;
 
-                // Safety:
-                // This mem transmute is extending the lifetime of the returned value.
-                // This is necessary because the rust borrow checker is too restrictive in giving out mutable references.
-                // It thinks the iterator can give out a mutable reference, while it's also possible to mutably borrow
-                // `obj` in the RingBufferMutIterator struct. This is however *never* possible because it's a private field
-                // Unfortunately this is a limitation of the rust compiler. It's well explained here:
-                // http://smallcultfollowing.com/babysteps/blog/2013/10/24/iterators-yielding-mutable-references/
-                unsafe { core::mem::transmute::<Option<&'_ mut T>, Option<&'rb mut T>>(res) }
+                res
             } else {
                 None
             }
@@ -236,22 +229,32 @@ pub use iter::{RingBufferIterator, RingBufferMutIterator};
 /// Implement the get, get_mut, get_absolute and get_absolute_mut functions on implementors
 /// of RingBuffer. This is to avoid duplicate code.
 macro_rules! impl_ringbuffer {
-    ($buf: ident, $index: ident) => {
+    ($get_unchecked: ident, $get_unchecked_mut: ident, $readptr: ident, $writeptr: ident, $mask: expr) => {
         #[inline]
         fn get(&self, index: isize) -> Option<&T> {
-            if self.len() > 0 {
-                let index = (index + self.$index as isize).rem_euclid(self.len() as isize) as usize;
-                self.$buf.get(index)
+            if !self.is_empty() {
+                let index = (self.$readptr as isize + index) as usize % self.len();
+
+                unsafe {
+                    // SAFETY: index has been modulo-ed and offset from readptr
+                    // to be within bounds
+                    Some(self.$get_unchecked(index))
+                }
             } else {
                 None
             }
         }
 
         #[inline]
-        unsafe fn get_mut_impl(&mut self, index: isize) -> Option<&mut T> {
-            if self.len() > 0 {
-                let index = (index + self.$index as isize).rem_euclid(self.len() as isize) as usize;
-                self.$buf.get_mut(index)
+        fn get_mut(&mut self, index: isize) -> Option<&mut T> {
+            if !self.is_empty() {
+                let index = (self.$readptr as isize + index) as usize % self.len();
+
+                unsafe {
+                    // SAFETY: index has been modulo-ed and offset from readptr
+                    // to be within bounds
+                    Some(self.$get_unchecked_mut(index))
+                }
             } else {
                 None
             }
@@ -259,8 +262,13 @@ macro_rules! impl_ringbuffer {
 
         #[inline]
         fn get_absolute(&self, index: usize) -> Option<&T> {
-            if index < self.len() {
-                self.$buf.get(index)
+            let read = $mask(self, self.$readptr);
+            let write = $mask(self, self.$writeptr);
+            if index >= read && index < write {
+                unsafe {
+                    // SAFETY: index has been checked against $mask to be within bounds
+                    Some(self.$get_unchecked(index))
+                }
             } else {
                 None
             }
@@ -268,11 +276,30 @@ macro_rules! impl_ringbuffer {
 
         #[inline]
         fn get_absolute_mut(&mut self, index: usize) -> Option<&mut T> {
-            if index < self.len() {
-                self.$buf.get_mut(index)
+            if index >= $mask(self, self.$readptr) && index < $mask(self, self.$writeptr) {
+                unsafe {
+                    // SAFETY: index has been checked against $mask to be within bounds
+                    Some(self.$get_unchecked_mut(index))
+                }
             } else {
                 None
             }
         }
+
+        #[inline]
+        fn len(&self) -> usize {
+            self.$writeptr - self.$readptr
+        }
+
+        #[inline]
+        fn clear(&mut self) {
+            self.$readptr = 0;
+            self.$writeptr = 0;
+        }
+
+        #[inline]
+        fn skip(&mut self) {
+            self.readptr += 1;
+        }
     };
 }
